import { ItemStack, system, world } from "@minecraft/server";
import { ActionFormData } from "@minecraft/server-ui";

world.afterEvents.playerSpawn.subscribe(({ player }) => {
  const hasJoined = player.getDynamicProperty("wm_vc:has_joined") == true;
  if (hasJoined) return;
  player.setDynamicProperty("wm_vc:has_joined", true);
  const inv = player.getComponent("inventory").container;
  if (inv.emptySlotsCount > 0)
    inv.addItem(new ItemStack("wm_vc:guide_book", 1));
  else
    player.dimension.spawnItem(
      new ItemStack("wm_vc:guide_book", 1),
      player.location
    );
});

world.afterEvents.playerBreakBlock.subscribe(
  ({
    brokenBlockPermutation,
    block,
    player,
    itemStackBeforeBreak: itemStack,
  }) => {
    try {
      if (!itemStack) return;
      if (!player.isSneaking) return;
      if (player.getGameMode() === "creative") return;
      const brokenBlockTypeId = brokenBlockPermutation.getItemStack().typeId
      const isPickaxe = itemStack.typeId.includes("_pickaxe");
      const isAxe = itemStack.typeId.includes("_axe");
      const hasTreeCapitator = itemStack.getDynamicProperty(
        "wm_vc:tree_capitator_scroll"
      );
      const hasVeinMiner = itemStack.getDynamicProperty(
        "wm_vc:vein_miner_scroll"
      );
      if (
        isPickaxe &&
        !Object.keys(ORE_BLOCKS).includes(brokenBlockTypeId)
      )
        return;
      if (
        isAxe &&
        !Object.keys(TREE_BLOCKS).includes(brokenBlockTypeId)
      )
        return;

      if (isPickaxe && !hasVeinMiner) return;
      if (isAxe && !hasTreeCapitator) return;
      breakArea(
        block,
        brokenBlockTypeId,
        itemStack,
        isAxe ? TREE_BLOCKS : ORE_BLOCKS
      );
    } catch (e) {}
  }
);

export function breakArea(
  centerBlock,
  centerBlockTypeId,
  itemStack,
  BLOCKTYPE
) {
  const enchants = itemStack.getComponent("enchantable");
  const hasSilkTouch = enchants.getEnchantment("minecraft:silk_touch");
  const hasFortune = enchants.getEnchantment("minecraft:fortune");
  const dimension = centerBlock.dimension;
  const adjacentBlocks = findAdjacentBlocks(centerBlock, centerBlockTypeId);
  try {
    for (const blockLocation of adjacentBlocks) {
      const currentBlock = dimension.getBlock(blockLocation);
      const foundBlock = BLOCKTYPE[currentBlock.typeId];
      if (!foundBlock) continue;
      if (currentBlock && currentBlock.typeId === centerBlockTypeId) {
        const randomAmount = random(foundBlock[1][0], foundBlock[1][1]) * (hasFortune?.level > 0 && Object.keys(ORE_BLOCKS).includes(centerBlockTypeId) ? hasFortune?.level + 1 : 1)
        currentBlock.dimension.spawnItem(
          new ItemStack(
            hasSilkTouch?.level > 0 ? currentBlock.typeId : foundBlock[0],
            randomAmount
          ),
          currentBlock.location
        );
        currentBlock.setType("minecraft:air");
        if (hasFortune && Object.keys(ORE_BLOCKS).includes(centerBlockTypeId)) {
          for (let i = 1; i < randomAmount; i++) currentBlock.dimension.spawnEntity('xp_orb', currentBlock.location)
          }
      }
    }
  } catch (e) {
    console.warn(e, e.stack);
  }

  return adjacentBlocks.length;
}

function random(min, max) {
  return Math.floor(Math.random() * (max - min + 1)) + min;
}

function findAdjacentBlocks(centerBlock, centerBlockTypeId, radius = 5) {
  const blocksToCheck = [centerBlock.location];
  const checkedLocations = new Set();
  const adjacentBlocks = [];

  while (blocksToCheck.length > 0) {
    const currentLocation = blocksToCheck.pop();
    const locationKey = `${currentLocation.x},${currentLocation.y},${currentLocation.z}`;

    if (checkedLocations.has(locationKey)) continue;

    checkedLocations.add(locationKey);

    const currentBlock = centerBlock.dimension.getBlock(currentLocation);
    if (
      currentBlock &&
      (currentBlock.typeId === centerBlockTypeId ||
        JSON.stringify(centerBlock.location) ===
          JSON.stringify(currentLocation))
    ) {
      adjacentBlocks.push(currentLocation);

      for (let dx = -1; dx <= 1; dx++) {
        for (let dy = -1; dy <= 1; dy++) {
          for (let dz = -1; dz <= 1; dz++) {
            if (dx === 0 && dy === 0 && dz === 0) continue;

            const adjacentLocation = {
              x: currentLocation.x + dx,
              y: currentLocation.y + dy,
              z: currentLocation.z + dz,
            };

            if (
              Math.abs(adjacentLocation.x - centerBlock.location.x) <= radius &&
              Math.abs(adjacentLocation.y - centerBlock.location.y) <=
                radius + 20 &&
              Math.abs(adjacentLocation.z - centerBlock.location.z) <= radius
            ) {
              blocksToCheck.push(adjacentLocation);
            }
          }
        }
      }
    }
  }

  return adjacentBlocks;
}

const TREE_BLOCKS = {
  "minecraft:stripped_oak_log": ["minecraft:stripped_oak_log", [1, 1]],
  "minecraft:stripped_pale_oak_log": ["minecraft:stripped_pale_oak_log", [1, 1]],
  "minecraft:stripped_dark_oak_log": ["minecraft:stripped_dark_oak_log",[1, 1]],
  "minecraft:stripped_birch_log": ["minecraft:stripped_birch_log", [1, 1]],
  "minecraft:stripped_spruce_log": ["minecraft:stripped_spruce_log", [1, 1]],
  "minecraft:stripped_acacia_log": ["minecraft:stripped_acacia_log", [1, 1]],
  "minecraft:stripped_jungle_log": ["minecraft:stripped_jungle_log", [1, 1]],
  "minecraft:stripped_cherry_log": ["minecraft:stripped_cherry_log", [1, 1]],
  "minecraft:stripped_mangrove_log": [ "minecraft:stripped_mangrove_log", [1, 1]],
  "minecraft:oak_log": ["minecraft:oak_log", [1, 1]],
  "minecraft:pale_oak_log": ["minecraft:pale_oak_log", [1, 1]],
  "minecraft:dark_oak_log": ["minecraft:dark_oak_log", [1, 1]],
  "minecraft:birch_log": ["minecraft:birch_log", [1, 1]],
  "minecraft:spruce_log": ["minecraft:spruce_log", [1, 1]],
  "minecraft:acacia_log": ["minecraft:acacia_log", [1, 1]],
  "minecraft:jungle_log": ["minecraft:jungle_log", [1, 1]],
  "minecraft:cherry_log": ["minecraft:cherry_log", [1, 1]],
  "minecraft:mangrove_log": ["minecraft:mangrove_log", [1, 1]],
};

const ORE_BLOCKS = {
  "minecraft:coal_ore": ["minecraft:coal", [1, 1]],
  "minecraft:copper_ore": ["minecraft:raw_copper", [2, 5]],
  "minecraft:iron_ore": ["minecraft:raw_iron", [1, 1]],
  "minecraft:gold_ore": ["minecraft:raw_gold", [1, 1]],
  "minecraft:lapis_ore": ["minecraft:lapis_lazuli", [4, 9]],
  "minecraft:diamond_ore": ["minecraft:diamond", [1, 1]],
  "minecraft:emerald_ore": ["minecraft:emerald", [1, 1]],
  "minecraft:nether_quartz_ore": ["minecraft:quartz", [1, 1]],
  "minecraft:redstone_ore": ["minecraft:redstone", [4, 5]],
  "minecraft:nether_gold_ore": ["minecraft:gold_nugget", [2, 6]],
  "minecraft:ancient_debris": ["minecraft:ancient_debris", [1, 1]],

  "minecraft:deepslate_coal_ore": ["minecraft:coal", [1, 1]],
  "minecraft:deepslate_copper_ore": ["minecraft:raw_copper", [2, 5]],
  "minecraft:deepslate_iron_ore": ["minecraft:raw_iron", [1, 1]],
  "minecraft:deepslate_gold_ore": ["minecraft:raw_gold", [1, 1]],
  "minecraft:deepslate_lapis_ore": ["minecraft:lapis_lazuli", [4, 9]],
  "minecraft:deepslate_diamond_ore": ["minecraft:diamond", [1, 1]],
  "minecraft:deepslate_emerald_ore": ["minecraft:emerald", [1, 1]],
  "minecraft:deepslate_redstone_ore": ["minecraft:redstone", [4, 5]],
};

world.afterEvents.itemUse.subscribe(({ source, itemStack }) => {
  let type = "";
  if (itemStack.typeId == "wm_vc:guide_book") return guideBook(source);
  if (itemStack.typeId == "wm_vc:tree_capitator_scroll") type = "_axe";
  else if (itemStack.typeId == "wm_vc:vein_miner_scroll") type = "_pickaxe";
  if (type == "") return;

  const allItems = getAllItemsByContains(source, type);
  pickItemMenu(source, allItems, itemStack.typeId);
});

function getAllItemsByContains(player, string) {
  if (!string) return [];
  const inv = player.getComponent("inventory").container;
  const matchingItems = [];
  for (let i = 0; i < inv.size; i++) {
    const item = inv.getItem(i);
    if (!item) continue;
    if (item.maxAmount > 1) continue;
    if (item.typeId.toLowerCase().includes(string.toLowerCase())) {
      matchingItems.push([item, i]);
    }
  }

  return matchingItems;
}

const getNameFromId = (object) => {
  return object.nameTag !== "" && object.nameTag !== undefined
    ? object.nameTag
    : object.typeId
        .split(":")[1]
        .replace(/_/g, " ")
        .replace(
          /\w\S*/g,
          (txt) => txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase()
        );
};

function pickItemMenu(player, items, upgradeId) {
  const f = new ActionFormData();
  f.title("Item Selector");
  f.body("Which item do you want to add the enchantment to?");
  const buttons = [];
  items.filter(item => item[0].getDynamicProperty(upgradeId) == undefined).forEach((item) => {
    buttons.push([
      `${getNameFromId(item[0])}\n${getSlotNames(item[1])} | Slot: ${
        item[1] + 1
      }`,
      "textures/items/" + getItemIcon(item[0].typeId),
    ]);
  });
  if (buttons.length == 0)
    return player.sendMessage(
      "§7[§c!§7] No §citems §7to add this §cscroll §7to."
    );
  buttons.forEach((button) => {
    f.button(button[0], button[1]);
  });
  f.show(player).then((r) => {
    if (r.canceled) return;
    const itemSelected = items.filter(item => item[0].getDynamicProperty(upgradeId) == undefined)[r.selection];
    const inv = player.getComponent("inventory").container;
    const foundItem = inv.getItem(itemSelected[1]);
    if (foundItem.typeId !== itemSelected[0].typeId) return;
    inv.setItem(player.selectedSlotIndex);
    foundItem.setDynamicProperty(upgradeId, true);
    const lore = foundItem.getLore();
    const newLore = [
      ...lore.splice(19),
      `§r§6§l${getLoreFromId(upgradeId)} §e[Shift to Enable]§r`,
    ];
    foundItem.setLore(newLore);
    inv.setItem(itemSelected[1], foundItem);
  });
}

function getLoreFromId(upgradeId) {
  switch (upgradeId) {
    case "wm_vc:tree_capitator_scroll":
      return "Tree Capitator";
    case "wm_vc:vein_miner_scroll":
      return "Vein Miner";
  }
}
function getItemIcon(itemId) {
  switch (itemId.split(":")[1]) {
    case "wooden_pickaxe":
      return "wood_pickaxe";
    case "stone_pickaxe":
      return "stone_pickaxe";
    case "iron_pickaxe":
      return "iron_pickaxe";
    case "golden_pickaxe":
      return "gold_pickaxe";
    case "diamond_pickaxe":
      return "diamond_pickaxe";
    case "netherite_pickaxe":
      return "netherite_pickaxe";
    case "wooden_axe":
      return "wood_axe";
    case "stone_axe":
      return "stone_axe";
    case "iron_axe":
      return "iron_axe";
    case "golden_axe":
      return "gold_axe";
    case "diamond_axe":
      return "diamond_axe";
    case "netherite_axe":
      return "netherite_axe";
  }
}
function getSlotNames(slot) {
  let slotName = "Hotbar";
  switch (slot) {
    case slot < 9:
      return (slotName = "Hotbar");
    case slot >= 9 && slot < 18:
      return (slotName = "Top Row");
    case slot >= 18 && slot < 27:
      return (slotName = "Middle Row");
    case slot >= 27 && slot < 35:
      return (slotName = "Bottom Row");
  }
  return slotName;
}
const body = [
  "Welcome §e$player§r!",
  "Have you ever wanted to speed up your §6mining§r? well, now with §eWan's Vein-Capitator§r You will be able to do that by §6crafting our scrolls§r!",
  "",
  "There currently are §62 scrolls:§r",
  "§eTree Capitator Scroll",
  "§eVein Miner Scroll§r",
  "",
  "By Adding these scrolls you will be able to §6destroy all blocks §rwith just one use of your item!",
  "",
  "All you will need to do, is just §6shift and break the block§r, the blocks that are currently will be destoyed automatically are:",
  "§eAll Wood log types",
  "§eAll Ore block types",
];

function guideBook(player) {
  const f = new ActionFormData();
  f.title("Vein-Capitator Guide Book");
  f.body(body.join("\n").replace("$player", player.name));
  f.button("Okay!");
  f.show(player);
}